# encoding=utf_8
import struct
import math
from collections import defaultdict, namedtuple
from random import randint
from construct import Container
from array import array
from warnings import warn
import numpy as np

class IncompleteWarning(UserWarning):
    pass
warnedincomplete = set()
def incomplete(message=''):
    if message in warnedincomplete:
        return
    warn(message, IncompleteWarning, 2)
    warnedincomplete.add(message)

FLAG_LOOP           = 0x0001
FLAG_TRANSPARENCY   = 0x0002
FLAG_SHIMMER        = 0x0004
FLAG_REPEATLAST     = 0x0008
FLAG_MOTION         = 0x0010
FLAG_MOTION_LOOP    = 0x0020
FLAG_MOTION_BOUNCE  = 0x0040
FLAG_MOTION_POINT   = 0x0080
FLAG_WAVE           = 0x0100
FLAG_RANDOM         = 0x4000
FLAG_INITIALIZED    = 0x8000

TERM_FLAG_MASK      = ~(FLAG_LOOP|FLAG_RANDOM)

STATE_1     = 1 # wait for buttondown or non-escape or -skip keypress
STATE_2     = 2 # wait for a period of time
STATE_3     = 3 # wait for a particular layer to finish animation
STATE_4     = 4 # wait for a fade to completed/started? - also cleared in map movement...
STATE_5     = 5 # waits for things like fades and layer clears clears to be applied
STATE_6     = 6 # waits for a particular layer to finish motion
STATE_7     = 7 # waits for a sound to be deleted?
STATE_8     = 8 # wait for text to be displayed for a period of time based on its length
STATE_9     = 9 # set when painting choices - not explicitly unset anywhere though. clearChoices does unset it among others
STATE_10    = 10 # wait for directional input - possibly only keys
STATE_11    = 11 # wait for there to be no keyboard or mouse input
STATE_12    = 12 # wait for the map movement to be finished
STATE_13    = 13 # wait for a button/key to be pressed (or a particular layer to be deleted)
STATE_14    = 14 # wait for a name to be entered
STATE_15    = 15 # wait for the character to be oriented?

INPUT_UP        = 0x01
INPUT_DOWN      = 0x02
INPUT_LEFT      = 0x04
INPUT_RIGHT     = 0x08
INPUT_ESCAPE    = 0x10
INPUT_SKIP      = 0x20

CHAR_TEXTSUBST = u'\1'
CHAR_TEXTCLEAR = u'＾' # 0x814F \uFF3E clear text
CHAR_TEXTCRLF  = u'￥' # 0x818F \uFFE5 line break
CHAR_TEXTHERO  = u'＄' # 0x8190 \uFF04 ＄ Replace with player name (default: 中田)

DEFAULT_HERO = u'中田'

SYSTEM_COLORS_START = (
    (0x00, 0x00, 0x00),
    (0x80, 0x00, 0x00),
    (0x00, 0x80, 0x00),
    (0x80, 0x80, 0x00),
    (0x00, 0x00, 0x80),
    (0x80, 0x00, 0x80),
    (0x00, 0x80, 0x80),
    (0xC0, 0xC0, 0xC0),
    (0xC0, 0xDC, 0xC0),
    (0xA4, 0xC8, 0xF0))
SYSTEM_COLORS_END = (
    (0xFF, 0xFB, 0xF0),
    (0xA0, 0xA0, 0xA4),
    (0x80, 0x80, 0x80),
    (0xFF, 0x00, 0x00),
    (0x00, 0xFF, 0x00),
    (0xFF, 0xFF, 0x00),
    (0x00, 0x00, 0xFF),
    (0xFF, 0x00, 0xFF),
    (0x00, 0xFF, 0xFF),
    (0xFF, 0xFF, 0xFF))

MAP_KEY_XYO = {
    0b0010: ( 0,  7, 0),
    0b1010: ( 5,  5, 0),
    0b1000: ( 7,  0, 3),
    0b1001: ( 5, -5, 1),
    0b0001: ( 0, -7, 1),
    0b0101: (-5, -5, 1),
    0b0100: (-7,  0, 2),
    0b0110: (-5,  5, 0),
}

class Interpreter(object):
    def __init__(self, ui, store):
        self.ui = ui
        self.store = store
        self.functions = dict((int(meth.split('_')[2], 16), getattr(self, meth)) for meth in dir(self) if meth.startswith('_op_'))
        self.debug = None
        self.drawing_x2 = 0
        self.drawing_y2 = 0
        self.ops = []
        self.index = 0
        self.loaded_frames = [None] * 8
        # unused update rectangles
        self.anim_layers = [AnimationLayer() for i in xrange(32)]
        self.track_names = [''] * 40
        self.text_lines = [''] # 25 originally
        self.names = [u''] * 40
        self.globals = np.zeros(256, dtype=np.int16)
        self.lines_i18n = [] # [0] * 25
        # registry entries...
        self.state = 0
        self.callstack = []
        self.palette = PaletteManager(ui)
        self.waitforlayer = 0
        self.pause_anim = False
        self.future_pause = False
        self.delay_countdown = 0
        self.typing_var = 0
        self.reading_countdown = 0
        self.reading_factor = 80 # reasonable default
        self.typing_layer = 0
        self.drawing_x = 0
        self.drawing_y = 0
        self.drawing_w = 640
        self.drawing_h = 400
        self.skip_type =0
        self.inputflags = 0
        self.buttonflags = 0
        self.input_var = 0
        self.inputshift = 0
        self.bytes_printed = 0
        self.continuehandler = None
        self.escapehandler = None
        self.exithandler = None
        self.loadhandler = None
        self.repeathandler = None
        self.mouse_caret = 0
        self.mouse_moved = False
        self.synth_mouse_moved = False
        self.mouse_pos_old = None
        self.mouse_pos_new = (0, 0)
        self.mouse_scrolling = False
        self.pause_anim = False
        self.i15dstrings = []
        self.dialogue_line = u''
        self.text_char_x = 0
        self.text_char_y = 0
        self.text_char_w = 640
        self.text_char_h = 80
        self.text_x = 0
        self.text_y = 400
        self.text_color = (255, 255, 255)
        self.font_height = 16 # reasonable default
        self.text_nlpadding = 0
        self.hero_name = DEFAULT_HERO
        self.caret_pos = (640 - 20, 480 - 20) # resonable default
        self.caret_type = 0
        self.caret_countdown = 0
        self.caret_index = 0
        self.choices_count = 0
        self.choices_var = 0
        self.choices_layout = 0
        self.choices_x = 0
        self.choices_y = 0
        self.choices_w = 0
        self.choices_h = 0
        self.choices_cellw = 0
        self.choices_text = []
        self.choiceidxs = []
        self.i15dchoices = False
        self.current_sound = None
        self.disabled_pcms = 0
        self.music_volume = 0x8000
        self.voice_volume = 0x8000
        self.music_stepto = 0xFFFF
        self.music_volume2 = 0xFFFF
        self.music_delta = 0
        self.last_trackn = 99
        self.noskipsound = 0
        self.use_dlgpop = 0
        self.dlgpop_bmp = None
        self.dlgpop_x = 0
        self.dlgpop_y = 384
        self.dlgpop_w = 640
        self.dlgpop_h = 80
        self.mapmove_orient = 0
        self.walk_sprite_index = 0
        self.walk_index = 0
        self.mapmove_x = 0
        self.mapmove_y = 0
        self.map_width = 0
        self.map_height = 0
        self.trail_index = 0
        self.move_trail = np.zeros((64, 3), dtype=np.int32)
        self.shimmer = Shimmer(0)
        self.translations = {}
    def nextstate(self):
        debug = self.debug
        if debug is not None: debug.nostate(self)
        funcs = self.functions
        while not self.state:
            op = self.nextop()
            if debug is not None: debug.nextop(self, op)
            try:
                # print '%4X %02x %d'%(op.address, op.instruction, self.skip_type)
                func = funcs[op.instruction]
            except KeyError:
                self.missingop(op)
            else:
                func(op)
        if debug is not None: debug.nextstate(self, self.state)
        # print self.state
    def nextop(self):
        i = self.index
        op = self.ops[i]
        self.index = i + 1
        # print '%04x'%op.address, op.fourcc
        return op
    def setops(self, ops):
        self.ops = ops
        self.addresses = dict((op.address, i) for i, op in enumerate(ops))
        self.index = 0
    def setaddress(self, address):
        self.index = self.addresses[address]
    def missingop(self, op):
        raise NotImplementedError('unimplemented op', op)

    def _op_00_terminate(self, op):
        raise RuntimeError('encountered terminator')
    def _op_01(self, op):
        self.drawing_x2 = op.x
        self.drawing_y2 = op.y
    def _op_02(self, op):
        self.drawing_x = self.drawing_x2
        self.drawing_y = self.drawing_y2
        self.drawing_w = op.w
        self.drawing_h = op.h
    def _op_0A_loadseq(self, op):
        self.anim_layers[op.a].resetseq(
            self.drawing_x2 + op.b, self.drawing_y2 + op.c,
            op.d & 0xFFFF, op.d >> 16 & 0xFF, op.d >> 24 & 0xFF,
            [SequenceItem(e.frame, e.b) for e in op.e])
        self.palette.resetpause()
    def _op_0B_rmlayer(self, op):
        self.anim_layers[op.level].flags = 0
        self.state = STATE_5
    def _op_4F(self, op):
        anim_layer = self.anim_layers[op.level]
        if anim_layer.flags & FLAG_INITIALIZED:
            anim_layer.flags &= TERM_FLAG_MASK
            self.future_pause = True
            self.waitforlayer = op.level
            self.state = STATE_3
    def _op_0C(self, op):
        anim_layer = self.anim_layers[op.level]
        if anim_layer.flags & FLAG_INITIALIZED:
            anim_layer.flags &= TERM_FLAG_MASK
            self.waitforlayer = op.level
            self.state = STATE_3
    def _op_0D(self, op):
        for anim_layer in self.anim_layers:
            anim_layer.flags = 0
        self.state = STATE_5
    def _op_0E(self, op):
        if not self.skip_type:
            self.caret_type = 2 if self.mouse_caret else 1
            self.caret_index = 0
            self.caret_countdown = 0
            self.state = STATE_1
    _op_0F = _op_0E
    def _op_13_delay(self, op):
        self.delay_countdown = n = 24 * op.a / 100
        if n and not self.skip_type:
            self.state = STATE_2
    def _op_17_tweenmove(self, op):
        self.anim_layers[op.a].resetvector(
            op.b, op.c, op.d, op.e, op.f, op.g, op.h)
    def _op_18(self, op):
        flags = FLAG_INITIALIZED | FLAG_MOTION
        if self.anim_layers[op.a].flags & flags == flags:
            self.waitforlayer = op.a
            self.state = STATE_6
    def _op_5C(self, op):
        self.anim_layers[op.a].stopvector()
    def _op_3D(self, op):
        self.ui.cleararea(op.x, op.y, op.w, op.h)
        self.state = STATE_5
    def _op_42(self, op):
        if not self.skip_type and self.reading_countdown:
            self.state = STATE_8
    def _op_43(self, op):
        self.reading_factor = op.a
    def _op_44(self, op):
        self.input_var = op.a
        self.state = STATE_10
    def _op_45(self, op):
        self.skip_type = 0
        self.globals[op.a] = self.inputflags
    def _op_46(self, op):
        if self.inputflags:
            self.state = STATE_11
    def _op_47(self, op):
        self.reset1()
    def reset1(self):
        del self.callstack[:]
        self.bytes_printed = 0
        for anim_layer in anim_layers:
            anim_layer.flags = 0
        self.loaded_frames[:] = [None] * len(self.loaded_frames)
    def resetandnext(self):
        del self.callstack[:]
        self.bytes_printed = 0
        self.continuehandler = None
        self.escapehandler = None
        self.exithandler = None
        self.loadhandler = None
        self.repeathandler = None
        self.mouse_caret = 0
        self.pause_anim = False
        self.ui.clearchoices(True)
        self.ui.clearcaret()
        self.ui.clearstate()
    def _op_49(self, op):
        self.globals[op.storeat] = self.mouseposcolor(op.a, op.b - 1)
    def mouseposcolor(self, anm_index, image_index):
        color = 255
        if self.mouse_moved or self.synth_mouse_moved:
            image = self.loaded_frames[anm_index].offsets[image_index].frame.value
            x, y, w, h = image.x, image.y, image.w, image.h
            mx, my = self.mouse_pos_new
            if x <= mx and y <= my and mx < x + w and my < y + h:
                color = ord(image.data[(h-my+y-1)*w+mx-x])
            self.mouse_moved = False
        return color
    def _op_4A_loadwin(self, op):
        del self.callstack[:]
        self.bytes_printed = 0
        self.continuehandler = None
        self.escapehandler = None
        self.exithandler = None
        self.loadhandler = None
        self.repeathandler = None
        self.ui.enableaction('exit', False)
        self.ui.enableaction('repeat', False)
        self.ui.enableaction('load', False)
        self.setops(self.store.loadwin(op.path))
        if self.debug is not None: self.debug.setcurrentwin(op.path, self.ops)
    def _op_4B(self, op):
        self.caret_pos = (op.x, op.y)
    def _op_4C(self, op):
        self.scroll_x1 = x1 = op.a
        self.scroll_x2 = x2 = op.b
        self.scroll_y1 = y1 = op.c
        self.scroll_y2 = y2 = op.d
        self.mouse_caret = n = op.b | op.c
        if n:
            flags = FLAG_INITIALIZED | FLAG_MOTION
            for anim_layer in self.anim_layers:
                if anim_layer.hasflags(flags):
                    anim_layer.resetscroll(x1, x2, y1, y2)
    def _op_4D(self, op):
        self.pause_anim = 0
        self.future_pause = 0
    def _op_4E(self, op):
        self.pause_anim = 1
    def _op_52(self, op):
        self.shimmer = Shimmer(op.a)
    def _op_56(self, op):
        self.loadglobals(op.a, 16, 256)
    def _op_57(self, op):
        self.saveglobals(op.a, 16, 256)
    def _op_59(self, op):
        self.anim_layers[op.a].resetwave(op.b, op.c, op.d, op.e, op.f)
    def _op_5A(self, op):
        flags = FLAG_INITIALIZED | FLAG_WAVE
        if self.anim_layers[op.a].flags & flags == flags:
            self.waitforlayer = op.a
            self.state = STATE_12
    def _op_5B(self, op):
        self.anim_layers[op.a].resetpoints([Point(c.a, c.b) for c in op.c])
    def _op_66(self, op):
        self.synth_mouse_moved = op.a
    def _op_67(self, op):
        self.ui.initi18n()
    def _op_6D(self, op):
        if self.anim_layers[op.a].flags:
            self.setaddress(op.target)
    def _op_79_startmap(self, op):
        self.mapvar = op.a
        self.state = STATE_15
    def _op_5F_nop(self, op):
        return

    def updatelayers(self, drawrect, saferegion):
        framebuffer = self.ui.framebuffer
        region = self.ui.frameregion
        for al_i, al in enumerate(self.anim_layers):
            if not al.flags & FLAG_INITIALIZED:
                if self.state == STATE_3 and self.waitforlayer == al_i:
                    self.ui.clearstate()
                    self.pause_anim = self.future_pause
                    self.future_pause = False
                if self.typing_layer == al_i and self.state == STATE_13:
                    self.globals[self.typing_var] = 0
                    self.ui.clearstate()
                    break
                continue

            al.frame_count -= 1
            while not al.frame_count:
                if al.flags & FLAG_RANDOM:
                    for i in xrange(randint(0, 3)):
                        al.sequence_index += 1
                        if al.sequence_index >= len(al.sequences):
                            al.sequence_index = 0

                i = al.sequence_index
                al.sequence_index = i + 1
                seq = al.sequences[i]
                al.frame_count = seq.count
                al.image_index = image_index = seq.index - 1
                if 0 <= image_index < 128:
                    loaded_frames = self.loaded_frames[al.anm_index]
                    image = withmask(loaded_frames.offsets[image_index].frame.value)

                    if al.flags & FLAG_WAVE:
                        x, y, w, h = al.sequence_x + image.x, al.sequence_y + image.y, al.vector_width, al.vector_height
                        region.adddims(x, y, w, h)
                        imgto = framebuffer[y:y+h, x:x+w]
                        x, y = al.vector_curx >> 16, al.vector_cury >> 16
                        imgfr = image.nparray[y:y+h]
                        xx = al.makewaveindex(x, h)
                        # TODO look into making this safer
                        for y, x in enumerate(xx):
                            imgto[y] = imgfr[y, x:x+w]
                        if al.advancewave() and self.state == STATE_12 and self.waitforlayer == al_i:
                            self.ui.clearstate()

                    else:
                        x, y, w, h, dx, dy = al.getoffsets(image.x, image.y, image.w, image.h)
                        viewrect, dx, dy = saferegion(x, y, w, h, dx, dy, image.w, image.h, drawrect)
                        x, y, w, h = viewrect

                        imgto = framebuffer[y:y+h, x:x+w]
                        x, y = dx, dy
                        imgfr = image.nparray[y:y+h, x:x+w]

                        # setup indexes (shimmer effect and optional)
                        index = Ellipsis
                        if al.flags & FLAG_SHIMMER:
                            index = self.shimmer.makeindex(w, h)

                            if al.flags & FLAG_TRANSPARENCY:
                                index &= image.masks[al.mask_color][y:y+h, x:x+w]

                        elif al.flags & FLAG_TRANSPARENCY:
                            index = image.masks[al.mask_color][y:y+h, x:x+w]

                        # copy
                        try:
                            imgto[index] = imgfr[index]

                        except Exception:
                            print_exc()
                            print >>sys.stderr, viewrect, dx, dy, framebuffer.shape, image.nparray.shape
                            print >>sys.stderr, al
                            al.flags = 0

                        else:
                            region.addrect(viewrect)

                if al.advancemotion() and self.state == STATE_6 and self.waitforlayer == al_i:
                    self.ui.clearstate()

            if al.frame_count == 1 and al.sequence_index >= len(al.sequences):
                if al.flags & FLAG_LOOP:
                    al.sequence_index = 0
                else:
                    if self.state == STATE_3 and self.waitforlayer == al_i:
                        self.ui.clearstate()
                        self.pause_anim = self.future_pause
                        self.future_pause = 0
                    if al.flags & FLAG_REPEATLAST:
                        al.sequence_index -= 1
                    else:
                        al.flags = 0

    def drawmap(self):
        mgl = self.globals[self.mapvar:]
        self.mapmove_orient = mgl[15]
        if mgl[0] != 1:
            if mgl[0] < 3:
                if mgl[0] == 0:
                    mgl[0] = 1
                img = self.loaded_frames[mgl[1]]
                img = img.offsets[0].frame.value
                self.map_width = img.w
                self.map_height = img.h
                self.walk_index = 0
                self.mapmove_x = 0
                self.mapmove_y = 0
                self.walk_sprite_index = 0
                x, y = mgl[4:6]
                for i in xrange(64):
                    move_trail = self.move_trail[63 - i]
                    if self.mapmove_orient == 0:
                        move_trail[0:2] = x, y - i
                    elif self.mapmove_orient == 1:
                        move_trail[0:2] = x, y + i
                    elif self.mapmove_orient == 2:
                        move_trail[0:2] = x + i, y
                    elif self.mapmove_orient == 3:
                        move_trail[0:2] = x - i, y
                    move_trail[2] = 3 * self.mapmove_orient
                self.trail_index = 0
                moved = True
            elif mgl[0] == 4:
                if mgl[17]:
                    mgl[17] -= 1
                    self.mapmove_orient = mgl[16]
                    if self.mapmove_orient == 0:
                        self.mapmove_orient, self.mapmove_x, self.mapmove_y = 1, 0, -7
                    elif self.mapmove_orient == 1:
                        self.mapmove_orient, self.mapmove_x, self.mapmove_y = 2, -7, 0
                    elif self.mapmove_orient == 2:
                        self.mapmove_orient, self.mapmove_x, self.mapmove_y = 0, 0, 7
                    elif self.mapmove_orient == 3:
                        self.mapmove_orient, self.mapmove_x, self.mapmove_y = 3, 7, 0
                    self.walk_index += 1
                    if self.walk_index == 12:
                        self.walk_index = 0
                    self.walk_sprite_index = self.walk_index / 3
                    moved = True
                else:
                    self.ui.clearstate()

        elif self.inputflags & 0x2F:
            if self.buttonflags & 0x20:
                mx, my = self.mouse_pos_new
                y = float(mx - self.drawing_w / 2) # x and y are reversed...
                x = float(my - self.drawing_h / 2)
                if y + x == 0.0:
                    self.mapmove_x, self.mapmove_y = 0, 0
                    theta = 0
                else:
                    theta = math.atan2(y, x)
                    self.mapmove_x = int(math.sin(theta) * 8.0)
                    self.mapmove_y = int(math.cos(theta) * 8.0)
                theta = 2 * theta / math.pi + 2.5
                n = int(theta) % 4
                if n == 0:
                    self.mapmove_orient = 1
                elif n == 1:
                    self.mapmove_orient = 2
                elif n == 2:
                    self.mapmove_orient = 0
                elif n == 3:
                    self.mapmove_orient = 3
            else:
                temp = MAP_KEY_XYO.get(self.inputflags & 0xF, None)
                if temp is not None:
                    self.mapmove_x, self.mapmove_y, self.mapmove_orient = temp
            self.walk_index += 1
            if self.walk_index == 12:
                self.walk_index = 0
            self.walk_sprite_index = self.walk_index / 3
            moved = True
        elif self.walk_index:
            self.walk_index += 1
            if self.walk_index == 12:
                self.walk_index = 0
            self.walk_sprite_index = self.walk_index / 3
            if not self.walk_sprite_index or self.walk_sprite_index == 2:
                self.walk_index = 0
            moved = True
        else:
            self.mapmove_x, self.mapmove_y = 0, 0
            if mgl[12]:
                moved = True
            else:
                moved = False

        if moved:
            img = self.loaded_frames[mgl[2]].offsets[0].frame.value
            def colorat(dx, dy):
                x, y = mgl[4] + dx, mgl[5] + dy
                i = img.w * (img.h - y / 16 - 1) + x / 16
                # i = self.map_width * (self.map_height - y - 16) / 256 + x / 16
                return ord(img.data[i])
            color = colorat(self.mapmove_x, self.mapmove_y)
            if color != 12:
                if self.mapmove_x and colorat(self.mapmove_x, 0) == 12:
                    self.mapmove_y = 0
                elif self.mapmove_y and colorat(0, self.mapmove_y) == 12:
                    self.mapmove_x = 0
                else:
                    if color <= 10:
                        for i in xrange(1, 1 if self.mapmove_x else 8):
                            if colorat(i, self.mapmove_y) == 12:
                                self.mapmove_x, self.mapmove_y = 7, 0
                                break
                            if colorat(-i, self.mapmove_y) == 12:
                                self.mapmove_x, self.mapmove_y = -7, 0
                                break
                        else:
                            for j in xrange(1, 1 if self.mapmove_y else 8):
                                if colorat(self.mapmove_x, j) == 12:
                                    self.mapmove_x, self.mapmove_y = 0, 7
                                    break
                                if colorat(self.mapmove_x, -j) == 12:
                                    self.mapmove_x, self.mapmove_y = 0, -7
                                    break
                            else:
                                self.mapmove_x, self.mapmove_y = 0, 0
                    else:
                        self.mapmove_x, self.mapmove_y = 0, 0
                if colorat(self.mapmove_x, self.mapmove_y) != 12:
                    self.mapmove_x, self.mapmove_y = 0, 0
            mgl[4] += self.mapmove_x
            mgl[5] += self.mapmove_y
            if color != 10:
                if color != 12:
                    mgl[0] = color
                    self.ui.clearstate()
                elif not self.walk_index and mgl[0] != 2 and not self.inputshift and randint(0, 99) < mgl[6]:
                    mgl[0] = 3
                    self.ui.clearstate()
            img = withmask(self.loaded_frames[mgl[1]].offsets[0].frame.value)

            dx = mgl[4] - self.drawing_w / 2 # center the view over the character
            dy = mgl[5] - self.drawing_h / 2
            # TODO make safe
            x, y, w, h = self.drawing_x, self.drawing_y, self.drawing_w, self.drawing_h
            self.ui.frameregion.adddims(x, y, w, h)
            imgto = self.ui.framebuffer[y:y+h, x:x+w]
            imgfr = img.nparray[dy:dy+h, dx:dx+w]
            imgto[...] = imgfr[...]

            if self.walk_sprite_index == 0:
                self.walk_sprite_index = 1
            elif self.walk_sprite_index == 1:
                self.walk_sprite_index = 0
            elif self.walk_sprite_index == 2:
                self.walk_sprite_index = 1
            elif self.walk_sprite_index == 3:
                self.walk_sprite_index = 2
            self.move_trail[self.trail_index][:] = mgl[4], mgl[5], 3 * self.mapmove_orient + self.walk_sprite_index
            characters = []
            for i in xrange(mgl[7]):
                j = (self.trail_index - 6 * i) % 64
                characters.append((
                    self.move_trail[j][0] - self.move_trail[self.trail_index][0],
                    self.move_trail[j][1] - self.move_trail[self.trail_index][1],
                    12 * mgl[8+i] + self.move_trail[j][2]))
            if self.mapmove_orient != 1:
                characters.reverse()
            characters.sort(key=lambda e: e[1])
            for x, y, i in characters:
                self.drawcharacter(x, y, i, mgl[3])
            if self.mapmove_x or self.mapmove_y:
                self.trail_index = (self.trail_index + 1) % 64

        if mgl[12]:
            self.palette.fade(True, (mgl[14], mgl[14], mgl[14]), mgl[13])
            mgl[12] = 0
            if mgl[0] == 2:
                self.state = 4
        elif mgl[0] == 2:
            self.ui.clearstate()
        mgl[15] = self.mapmove_orient

    def drawcharacter(self, x, y, j, i):
        img = withmask(self.loaded_frames[i].offsets[j].frame.value)
        # TODO make safe
        w, h = img.w, img.h
        x, y = (self.drawing_x + self.drawing_w / 2 + x - w / 2,
                self.drawing_y + self.drawing_h / 2 + y - h / 2)
        imgto = self.ui.framebuffer[y:y+h, x:x+w]
        x, y = 0, 0
        imgfr = img.nparray[y:y+h, x:x+w]
        mask = img.masks[0x0C][y:y+h, x:x+w]
        imgto[mask] = imgfr[mask]

    # graphic resources
    def _op_03_loadanm(self, op):
        self.loaded_frames[op.a] = self.store.loadanm(op.path)
    def _op_04_freeanm(self, op):
        self.loaded_frames[op.a] = None
    def _op_10(self, op):
        self.loaded_frames[op.a] = Container(uncompressed=True, frame_count=0, offset_start=772, offsets=[])
    def _op_12_copyimage(self, op):
        frames = self.loaded_frames[op.c]
        frameorig = self.loaded_frames[op.a].offsets[op.b-1].frame.value
        framecopy = Container(x=op.d, y=op.e, w=frameorig.w, h=frameorig.h, data=frameorig.data)
        frames.offsets.append(Container(frame=Container(value=framecopy)))
        frames.frame_count += 1
    def _op_61(self, op):
        frames = self.loaded_frames[op.a]
        w, h = op.w, op.h
        framecopy = Container(x=op.x, y=op.y, w=w, h=h, data='\0'*(w*h))
        frames.offsets.append(Container(frame=Container(value=framecopy)))
        frames.frame_count += 1

    # palette ops
    def _op_05_fadein(self, op):
        self.palette.fade(True, (op.a, op.a, op.a), op.div)
        self.state = STATE_4
    def _op_06_fadeout(self, op):
        self.palette.fade(False, (op.a, op.a, op.a), op.div)
        self.state = STATE_4
    def _op_07_loadpalette(self, op):
        self.palette.setpalette([(f.red, f.green, f.blue) for f in self.loaded_frames[op.a].palette])
    def _op_08(self, op):
        self.palette.pausefade()
        self.state = STATE_5
    def _op_09(self, op):
        self.palette.stopfade()
        self.state = STATE_5
    def _op_28(self, op):
        self.palette.fadetocolors(op.type, op.frames, [(b.i, b.r, b.g, b.b) for b in op.colors])
        # QUESTION no state change?
    def _op_3A(self, op):
        self.palette.fadetoblack(op.a)
        self.state = STATE_4
    def _op_41(self, op):
        # QUESTION assuming i got the order right...
        self.palette.setpalettebyindex([(b.i, b.r, b.g, b.b) for b in op.b])
    def _op_50(self, op):
        self.palette.fade(True, (op.r, op.g, op.b), op.div)
        self.state = STATE_4
    def _op_51(self, op):
        self.palette.fade(False, (op.r, op.g, op.b), op.div)
        self.state = STATE_4
    def _op_55(self, op):
        self.palette.setrange(op.a, op.b)
    def _op_64(self, op):
        self.palette.paletteshift([(a.frames, np.asarray([(b.i, b.r, b.g, b.b) for b in a.b], np.uint8)) for a in op.a])
        # QUESTION no state change?

    # audio
    def _op_19(self, op):
        self.music_volume2 = self.music_stepto
        self.music_delta = 0
        self.ui.setvolumes()
        self.ui.playbgm(op.sndnum, op.a)
    def _op_1A(self, op):
        self.music_delta = 0
        self.ui.stopbgm()
    def _op_1B(self, op): 
        self.track_names[op.sndnum] = op.path
    def _op_65(self, op):
        self.last_trackn = op.a
    def _op_5D(self, op):
        self.music_stepto = n = op.volume | (op.volume << 8)
        self.music_volume2 = n
        self.music_delta = 0
        self.ui.setvolumes()
        # print '5D', self.music_delta, self.music_stepto, self.music_volume2
    def _op_5E(self, op):
        self.music_delta = -0xFFFF / op.frames
        self.music_stepto = 0
        # print '5E', self.music_delta, self.music_stepto
    def _op_60(self, op):
        n = op.volume | (op.volume << 8)
        self.music_delta = (n - self.music_stepto) / op.frames
        self.music_stepto = n
        # print '60', self.music_delta, self.music_stepto
    def _op_1D(self, op):
        flags = op.flags
        if not flags & self.disabled_pcms:
            self.ui.loadpcm(op.path, flags & 1)
    def _op_1E(self, op):
        if self.current_sound is not None:
            self.state = STATE_7
            self.ui.deletepcm(1)
    def _op_1F(self, op):
        if self.current_sound is not None:
            self.state = STATE_7
            if self.skip_type == 1:
                self.ui.deletepcm(1)
    def _op_63(self, op):
        self.disabled_pcms = op.flags
    def _op_6B(self, op):
        self.ui.composepcm(op.path, 441 * op.delay, op.flags)
    def _op_6C(self, op):
        if not op.a & self.disabled_pcms:
            self.ui.startpcm(op.a & 1)
    def _op_78(self, op):
        self.noskipsound = op.disable
    def advancevolume(self):
        if not self.music_delta:
            return False
        self.music_volume2 += self.music_delta
        if self.music_delta <= 0:
            if self.music_volume2 < self.music_stepto:
                self.music_volume2 = self.music_stepto
                self.music_delta = 0
        else:
            if self.music_volume2 > self.music_stepto:
                self.music_volume2 = self.music_stepto
                self.music_delta = 0
        return True

    # dialogue
    def _op_20_textarea(self, op):
        self.text_x = op.x
        self.text_y = op.y
        self.text_char_x = 0
        self.text_char_y = 0
        self.text_char_w = op.w
        self.text_char_h = op.h
    def _op_21(self, op):
        self.bytes_printed = 0
        if op.address in self.translations:
            self.dialogue_line = self.translations[op.address]
        else:
            self.dialogue_line = op.text
        self.ui.drawdialogue()
    def _op_68(self, op):
        self.bytes_printed = 0
        self.lines_i18n.append(op.line)
        self.ui.drawdialoguei18n()
    def _op_22(self, op):
        self.text_nlpadding = op.a
    def _op_23(self, op):
        self.font_height = op.a
        self.ui.setfontheight(op.a)
    def _op_62(self, op):
        self.use_dlgpop = n = op.a
        if n == 2:
            self.ui.showdialoguepopup(True, update=True)
            self.dlgpop_shown = True
        else:
            self.ui.showdialoguepopup(False)
            if n == 3:
                self.dlgpop_shown = False
    def _op_24_name(self, op):
        if op.address in self.translations:
            self.names[op.name] = self.translations[op.address]
        else:
            self.names[op.name] = op.text
    def _op_76(self, op):
        self.ui.askforname()
        self.state = STATE_14
    def _op_7B(self, op):
        self.dlgpop_bmp = data = self.store.loadbmp(op.path)
        w, h = struct.unpack_from('II', data, 0x12)
        self.dlgpop_x = op.a
        self.dlgpop_y = op.b
        self.dlgpop_w, self.dlgpop_h = w, h

    def getdialogue(self, draw=True):
        # TODO find a way to return to accumulating characters in a list instead of concatenating to a string
        actions = []
        self = self
        x, y, w, h = self.text_char_x, self.text_char_y, self.text_char_w, self.text_char_h
        x_old, y_old = x, y
        line, text = self.dialogue_line, self.text_lines
        nlpadding, bytes_printed = self.text_nlpadding, self.bytes_printed

        i, j = 0, len(line)
        while i < j:
            c = line[i]
            i += 1
            if c == CHAR_TEXTSUBST:
                self.dialogue_line = self.names[ord(line[i])-1]
                i += 1
                self.text_char_x, self.text_char_y, self.bytes_printed = x, y, bytes_printed
                actions.extend(self.getdialogue(False))
                x, y, bytes_printed = self.text_char_x, self.text_char_y, self.bytes_printed
            elif c == CHAR_TEXTCLEAR:
                text[:] = ['']
                fh = self.font_height
                actions.append(('clear', (self.text_x, self.text_y, fh * w / 2, fh * h)))
                self.text_char_x = 0
                x_old = 0
                x = 0
                self.text_char_y = 0
                y_old = 0
                y = 0
                del self.lines_i18n[:]
            elif c == CHAR_TEXTCRLF:
                y += 1
                text.append('')
                x = 0
            elif c == CHAR_TEXTHERO:
                self.dialogue_line = self.hero_name
                self.text_char_x, self.text_char_y, self.bytes_printed = x, y, bytes_printed
                actions.extend(self.getdialogue(False))
                x, y, bytes_printed = self.text_char_x, self.text_char_y, self.bytes_printed
            else:
                if not x and draw:
                    text[y] += ' ' * nlpadding
                    x += nlpadding
                cw = 1 if ord(c) <= 0x80 else 2
                if x > w - cw:
                    y += 1
                    text.append(' ' * nlpadding)
                    x = nlpadding
                text[y] += c
                bytes_printed += cw
                x += cw

        if draw:
            tx, ty, fh = self.text_x, self.text_y, self.font_height
            args = []
            while y + 1 > y_old:                
                s = text[y_old] # ignoring x_old, going to draw the entire line over the prior draw.
                args.append((s, tx, ty + fh * y_old)) # x is supposed to be: tx + (fh * x_old) / 2
                y_old += 1
                x_old = 0
            actions.append(('draw', args))

            self.reading_countdown = self.reading_factor * 2 * bytes_printed / 100

        self.text_char_x = x
        self.text_char_y = y
        self.dialogue_line = ''
        return actions

    def getalldialogue(self):
        tx, ty, fh, text = self.text_x, self.text_y, self.font_height, self.text_lines
        return [(text[i], tx, ty + fh * i) for i in xrange(self.text_char_y + 1)]

    def getdialoguei18n(self, enc, i=-1):
        text = self.text_lines
        x, y, w, h = self.text_char_x, self.text_char_y, self.text_char_w, self.text_char_h
        tx, ty, fh = self.text_x, self.text_y, self.font_height
        x2, y2 = x, y
        self.dialogue_line = line = self.i15dstrings[self.lines_i18n[i]]
        encoded = False
        if enc == 'cp932':
            line = line.encode('cp932')
            encoded = True
        n = len(line)
        while n:
            linei = w - x
            if linei > n:
                linei = n
            elif not encoded:
                linej = linei = min(linei + 20, n)
                nospaces = 0
                while linej > 0:
                    if nospaces > 20:
                        linei = linej
                        break
                    linej -= 1
                    if line[linej] == u' ':
                        nospaces = 0
                        tw = self.ui.gettextwidth(text[y][:x] + line[:linej])
                        if tw < w * (fh / 2):
                            linei = linej + 1
                            break
                    nospaces += 1
            elif ord(line[linei - 2]) < 0x80 and ord(line[linei - 1]) > 0x80:
                linei -= 1
            s = line[:linei]
            if encoded: s = s.decode(enc)
            text[y] += s
            line = line[linei:]
            n -= linei
            if text[y][-1] == '\n':
                text[y] = text[y][:-1]
                y += 1
                x = 0
                text.append(u'')
            else:
                y += 1
                c = u'【' if enc == 'cp932' else u'['
                if text[0][0] == c:
                    x = 4
                    text.append(u'    ')
                else:
                    x = 0
                    text.append(u'')

        lines = []
        while y2 <= y:
            s = text[y2] # ignoring x2, going to draw the entire line over the prior draw.
            lines.append((s, tx, ty + fh * y2)) # x is supposed to be: tx + (fh * x2) / 2
            x2 = 0
            y2 += 1

        self.reading_countdown = self.reading_factor * 2 * self.bytes_printed / 100

        self.text_char_x = x
        self.text_char_y = y
        self.dialogue_line = ''
        return lines

    # user-influenced control flow
    def _op_15_setcontinue(self, op):
        self.continuehandler = op.target
    def _op_16_setescape(self, op):
        self.escapehandler = op.target
    def _op_25(self, op):
        self.callstack.append(self.ops[self.index].address)
        self.setaddress(op.target)
    def _op_26(self, op):
        self.setaddress(self.callstack.pop())
    def _op_27(self, op):
        self.ui.setfullscreen(op.a)
    def _op_29(self, op):
        self.exithandler = op.target
        self.ui.enableaction('exit')
    def _op_2A(self, op):
        self.exithandler = None
        self.ui.enableaction('exit', False)
    def _op_2B(self, op):
        self.loadhandler = op.target
        self.ui.enableaction('load')
    def _op_2C(self, op):
        self.loadhandler = None
        self.ui.enableaction('load', False)
    def _op_53(self, op):
        self.repeathandler = op.target
        self.ui.enableaction('repeat')
    def _op_54(self, op):
        self.repeathandler = None
        self.ui.enableaction('repeat', False)
    def _op_2D(self, op):
        self.continuehandler = None
        self.escapehandler = None
    def _op_40(self, op):
        self.enablesave = b = op.a
        self.ui.enableaction('save', b)

    # control flow and state (minut input storing)
    def _op_14_jump(self, op):
        self.setaddress(op.target)
    def _op_2E(self, op):
        if self.globals[op.a] == op.b:
            self.setaddress(op.target)
    def _op_2F(self, op):
        if self.globals[op.a] != op.b:
            self.setaddress(op.target)
    def _op_30(self, op):
        if self.globals[op.a] <= op.b:
            self.setaddress(op.target)
    def _op_31(self, op):
        if self.globals[op.a] >= op.b:
            self.setaddress(op.target)
    def _op_6F(self, op):
        if self.globals[op.a] == self.globals[op.b]:
            self.setaddress(op.target)
    def _op_70(self, op):
        if self.globals[op.a] != self.globals[op.b]:
            self.setaddress(op.target)
    def _op_71(self, op):
        if self.globals[op.a] <= self.globals[op.b]:
            self.setaddress(op.target)
    def _op_72(self, op):
        if self.globals[op.a] >= self.globals[op.b]:
            self.setaddress(op.target)
    def _op_32(self, op):
        self.switch_value = self.globals[op.a]
    def _op_33(self, op):
        if self.switch_value != op.a:
            self.setaddress(op.target)
    def _op_48(self, op):
        if not op.a & self.switch_value:
            self.setaddress(op.target)
    def _op_7A_caseand(self, op):
        if op.a & self.switch_value:
            self.setaddress(op.target)
    def _op_34(self, op):
        self.globals[op.a] = op.b
    def _op_35(self, op):
        self.globals[op.a] += 1
    def _op_36(self, op):
        self.globals[op.a] -= 1
    def _op_37(self, op):
        self.globals[op.a] += op.b
    def _op_38(self, op):
        self.globals[op.a] |= op.b
    def _op_39(self, op):
        self.globals[op.a] &= op.b
    def _op_58(self, op):
        self.globals[op.a] = self.globals[op.b]
    def _op_6A(self, op):
        self.globals[op.a] = randint(0, op.mod - 1)
    def _op_3B(self, op):
        self.loadglobals("SaveSystem", 0, 16)
    def _op_3C(self, op):
        self.saveglobals("SaveSystem", 0, 16)
    def loadglobals(self, name, start, end):
        data = self.store.loadglobals(name)
        if data is not None and len(data) == end - start:
            self.globals[start:end] = data
    def saveglobals(self, name, start, end):
        self.store.saveglobals(name, self.globals[start:end])

    # choices
    def _op_3E(self, op):
        self.skip_type = 0
        self.choices_var = op.a
        self.choices_x = op.b
        self.choices_y = op.c
        self.choices_w = op.d * self.font_height / 2
        self.choices_count = op.n
        self.choices_h = op.n * self.font_height
        if op.address in self.translations:
            ss = self.translations[op.address]
        else:
            ss = [f.text for f in op.f]
        self.choices_text = [self.names[ord(s[1])] if s[0] == '\1' else s for s in ss]
        self.ui.updatechoicelayout()
        self.ui.paintchoices(True)
        self.state = STATE_9
    def _op_69(self, op):
        self.skip_type = 0
        self.choices_var = op.a
        self.choices_x = op.b
        self.choices_y = op.c
        self.choices_count = op.d
        self.choices_h = op.d * self.font_height
        self.choiceidxs = indexes = [e.a for e in op.e]
        self.choices_text = [self.i15dstrings[n] for n in indexes]
        self.ui.updatechoicelayout()
        self.ui.paintchoices(True)
        self.state = STATE_9
        self.i15dchoices = True
    def _op_3F(self, op):
        self.choices_layout = op.a

    # typing
    def _op_6E(self, op):
        self.typing_var = op.a
        self.typing_layer = op.b
        self.state = STATE_13
    def _op_73(self, op): # this is probably used in Typing
        if op.address in self.translations:
            self.dialogue_line = self.translations[op.address]
        else:
            self.dialogue_line = op.a
        self.ui.drawtext()
    def _op_74(self, op):
        self.text_color = (op.r, op.g, op.b)
    def _op_75(self, op):
        self.text_position = (op.x, op.y)

class AnimationLayer(object):
    __slots__ = (
        'flags',
        'anm_index',
        'mask_color',
        'frame_count',
        'image_index',
        'sequence_index',
        'sequence_x',
        'sequence_y',
        'vector_curx',
        'vector_cury',
        'vector_width',
        'vector_height',
        'vector_deltax',
        'vector_deltay',
        'vector_countdown',
        'vector_initx',
        'vector_inity',
        'vector_initcount',
        'vector_endx',
        'vector_endy',
        'wave_cur',
        'wave_deltax',
        'wave_shift',
        'wave_delta',
        'wave_deltay',
        'wave_countdown',
        'point_countdown',
        'point_count',
        'current_point',
        'sequences', # 27 {short, short}[?]
        'points' # 539 {int, int}[256]
    )
    def __init__(self):
        for name in self.__slots__:
            setattr(self, name, 0)
        self.sequences = []
        self.points = []
    def __str__(self):
        return 'AnimationLayer({0})'.format(', '.join('{0}={1}'.format(name, getattr(self, name)) for name in self.__slots__))

    def hasflags(self, flags):
        return self.flags & flags == flags

    def resetseq(self, x, y, flags, anm, color, seqs):
        self.sequence_x = x
        self.sequence_y = y
        self.frame_count = 1
        self.sequence_index = 0
        self.flags = flags
        self.anm_index = anm
        self.mask_color = color
        self.sequences = seqs

    def resetvector(self, x1, x2, w, y1, y2, h, steps):
        x1, x2, y1, y2 = x1 << 16, x2 << 16, y1 << 16, y2 << 16
        self.vector_initx = x1
        self.vector_inity = y1
        self.vector_curx = x1
        self.vector_cury = y1
        self.vector_endx = x2
        self.vector_endy = y2
        self.vector_deltax = x2 - x1
        self.vector_deltay = y2 - y1
        self.vector_width = w
        self.vector_height = h
        self.vector_countdown = steps
        self.vector_initcount = steps
        if steps:
            self.vector_deltax /= steps
            self.vector_deltay /= steps
        else:
            self.vector_deltax = 0
            self.vector_deltay = 0

    def stopvector(self):
        self.vector_countdown = 0
        self.vector_initcount = 0

    def resetscroll(self, x1, x2, y1, y2):
        self.vector_deltax = self.vector_endx / x2 if x2 else 0
        self.vector_deltay = self.vector_endy / y2 if y2 else 0
        self.vector_curx = x1 * self.vector_deltax
        self.vector_cury = y1 * self.vector_deltay

    def resetwave(self, value, delta, deltax, deltay, steps):
        self.wave_cur = value
        self.wave_deltax = deltax # was * 8, but i 1) separate sine and cosine and 2) address them by int instead of char
        self.wave_delta = delta
        self.wave_deltay = deltay # see above
        self.wave_countdown = steps
        if steps:
            self.wave_delta = (delta - value) / steps

    def makewaveindex(self, x, h):
        xx = np.arange(h); xx *= self.wave_deltax; xx += self.wave_shift; xx &= 0xFF
        xx = COSINES[xx]; xx *= self.wave_cur; xx /= 10000; xx += x
        return xx

    def advancewave(self):
        self.wave_shift = (self.wave_shift + self.wave_deltay) & 0xFF # was 0x7ff, see comment in resetwave
        if self.wave_countdown:
            self.wave_countdown -= 1
            self.wave_cur += self.wave_delta
        else:
            return True
        return False

    def resetpoints(self, points):
        n = len(points)
        self.point_countdown = n
        self.point_count = n
        self.current_point = 0
        self.points = points

    def getoffsets(self, ix, iy, iw, ih):
        x, y = self.sequence_x + ix, self.sequence_y + iy
        if self.flags & FLAG_MOTION:
            w, h = self.vector_width, self.vector_height # and this is why tweens are still specified for points
            if self.flags & FLAG_MOTION_POINT:
                if self.points:
                    dx, dy = self.points[self.current_point]
                else:
                    dx, dy = 0, 0
            else:
                dx, dy = self.vector_curx >> 16, self.vector_cury >> 16
        else:
            w, h, dx, dy = iw, ih, 0, 0
        return x, y, w, h, dx, dy

    def advancemotion(self):
        if not self.flags & FLAG_MOTION:
            return False
        clear = False
        if self.flags & FLAG_MOTION_POINT:
            if self.point_countdown:
                self.point_countdown -= 1
                if self.point_countdown:
                    self.current_point += 1
                elif self.flags & FLAG_MOTION_LOOP:
                    self.point_countdown = self.point_count
                    self.current_point = 0
            else:
                clear = True
        elif self.vector_countdown:
            self.vector_curx += self.vector_deltax
            self.vector_cury += self.vector_deltay
            self.vector_countdown -= 1
            if not self.vector_countdown and self.flags & FLAG_MOTION_LOOP:
                if self.flags & FLAG_MOTION_BOUNCE:
                    self.vector_deltax = -self.vector_deltax
                    self.vector_deltay = -self.vector_deltay
                else:
                    self.vector_curx = self.vector_initx
                    self.vector_cury = self.vector_inity
                self.vector_countdown = self.vector_initcount
        else:
            clear = True
        return clear

SequenceItem = namedtuple('SequenceItem', 'index count')
Point = namedtuple('Point', 'x y')

SINES = np.array([
       0,    25,    49,    74,    98,   122,   147,   171,   195,
     219,   243,   267,   290,   314,   337,   360,   383,   405,
     428,   450,   471,   493,   514,   535,   556,   576,   596,
     615,   634,   653,   672,   690,   707,   724,   741,   757,
     773,   788,   803,   818,   831,   845,   858,   870,   882,
     893,   904,   914,   924,   933,   942,   950,   957,   964,
     970,   976,   981,   985,   989,   992,   995,   997,   999,
    1000,  1000,  1000,   999,   997,   995,   992,   989,   985,
     981,   976,   970,   964,   957,   950,   942,   933,   924,
     914,   904,   893,   882,   870,   858,   845,   831,   818,
     803,   788,   773,   757,   741,   724,   707,   690,   672,
     653,   634,   615,   596,   576,   556,   535,   514,   493,
     471,   450,   428,   405,   383,   360,   337,   314,   290,
     267,   243,   219,   195,   171,   147,   122,    98,    74,
      49,    25,     0,   -25,   -49,   -74,   -98,  -122,  -147,
    -171,  -195,  -219,  -243,  -267,  -290,  -314,  -337,  -360,
    -383,  -405,  -428,  -450,  -471,  -493,  -514,  -535,  -556,
    -576,  -596,  -615,  -634,  -653,  -672,  -690,  -707,  -724,
    -741,  -757,  -773,  -788,  -803,  -818,  -831,  -845,  -858,
    -870,  -882,  -893,  -904,  -914,  -924,  -933,  -942,  -950,
    -957,  -964,  -970,  -976,  -981,  -985,  -989,  -992,  -995,
    -997,  -999, -1000, -1000, -1000,  -999,  -997,  -995,  -992,
    -989,  -985,  -981,  -976,  -970,  -964,  -957,  -950,  -942,
    -933,  -924,  -914,  -904,  -893,  -882,  -870,  -858,  -845,
    -831,  -818,  -803,  -788,  -773,  -757,  -741,  -724,  -707,
    -690,  -672,  -653,  -634,  -615,  -596,  -576,  -556,  -535,
    -514,  -493,  -471,  -450,  -428,  -405,  -383,  -360,  -337,
    -314,  -290,  -267,  -243,  -219,  -195,  -171,  -147,  -122,
     -98,   -74,   -49,   -25])
COSINES = np.array([
    1000,  1000,   999,   997,   995,   992,   989,   985,   981,
     976,   970,   964,   957,   950,   942,   933,   924,   914,
     904,   893,   882,   870,   858,   845,   831,   818,   803,
     788,   773,   757,   741,   724,   707,   690,   672,   653,
     634,   615,   596,   576,   556,   535,   514,   493,   471,
     450,   428,   405,   383,   360,   337,   314,   290,   267,
     243,   219,   195,   171,   147,   122,    98,    74,    49,
      25,     0,   -25,   -49,   -74,   -98,  -122,  -147,  -171,
    -195,  -219,  -243,  -267,  -290,  -314,  -337,  -360,  -383,
    -405,  -428,  -450,  -471,  -493,  -514,  -535,  -556,  -576,
    -596,  -615,  -634,  -653,  -672,  -690,  -707,  -724,  -741,
    -757,  -773,  -788,  -803,  -818,  -831,  -845,  -858,  -870,
    -882,  -893,  -904,  -914,  -924,  -933,  -942,  -950,  -957,
    -964,  -970,  -976,  -981,  -985,  -989,  -992,  -995,  -997,
    -999, -1000, -1000, -1000,  -999,  -997,  -995,  -992,  -989,
    -985,  -981,  -976,  -970,  -964,  -957,  -950,  -942,  -933,
    -924,  -914,  -904,  -893,  -882,  -870,  -858,  -845,  -831,
    -818,  -803,  -788,  -773,  -757,  -741,  -724,  -707,  -690,
    -672,  -653,  -634,  -615,  -596,  -576,  -556,  -535,  -514,
    -493,  -471,  -450,  -428,  -405,  -383,  -360,  -337,  -314,
    -290,  -267,  -243,  -219,  -195,  -171,  -147,  -122,   -98,
     -74,   -49,   -25,     0,    25,    49,    74,    98,   122,
     147,   171,   195,   219,   243,   267,   290,   314,   337,
     360,   383,   405,   428,   450,   471,   493,   514,   535,
     556,   576,   596,   615,   634,   653,   672,   690,   707,
     724,   741,   757,   773,   788,   803,   818,   831,   845,
     858,   870,   882,   893,   904,   914,   924,   933,   942,
     950,   957,   964,   970,   976,   981,   985,   989,   992,
     995,   997,   999,  1000])

class PaletteManager(object):
    def __init__(self, ui):
        self.ui = ui
        self.bitmap_palette = np.zeros([256, 3], np.uint8)
        self.palette_from = 10
        self.palette_to = 246
        self.master_palette = np.zeros((256, 3), np.uint8)
        self.current_palette = np.zeros((256, 3), np.uint8)
        self.current_palette[:10] = SYSTEM_COLORS_START
        self.current_palette[246:] = SYSTEM_COLORS_END
        self.current_rgb = np.zeros((256, 3), np.int32)
        self.delta_rgb = np.zeros((256, 3), np.int32)
        self.fade_in = 0
        self.fade_color = 0
        self.palette_changed = False
        self.future_palette_changed = False

    def setrange(self, start, end):
        self.palette_from = start
        self.palette_to = end

    def setpalette(self, colors):
        colors = np.asarray(colors, np.uint8)
        a, b = self.palette_from, self.palette_to
        self.bitmap_palette[a:b] = colors[a:b]
        self.master_palette[a:b] = colors[a:b]
        self.endfade()

    def setpalettebyindex(self, colors):
        colors = np.asarray(colors, np.uint8)
        self.master_palette[colors[:, 0], :3] = colors[:, 1:]
        self.bitmap_palette[colors[:, 0]] = colors[:, 1:]
        self.endfade()

    def fade(self, fade_in, color, steps):
        self.fade_color = color
        self.fade_in = fade_in
        self.palchange_type = 1
        self.palchange_countdown = steps
        self.palette_changed = True

        master, curpal = self.master_palette, self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb
        a, b = self.palette_from, self.palette_to

        if fade_in:
            curpal[a:b] = color
            currgb[a:b] = curpal[a:b]
            delta[a:b] = master[a:b]
            delta[a:b] -= curpal[a:b]
        else:
            curpal[a:b] = master[a:b]
            currgb[a:b] = master[a:b]
            delta[a:b] = color
            delta[a:b] -= master[a:b]
        currgb[a:b] <<= 8
        delta[a:b] <<= 8
        delta[a:b] /= steps

    def fadetocolors(self, type, steps, colors):
        self.palchange_type = type
        self.palchange_count = steps
        self.palchange_countdown = steps
        self.palette_changed = True
        divisor = 1 if type == 3 else steps

        master, curpal = self.master_palette, self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb

        curpal[10:246] = master[10:246]
        currgb[10:246] = master[10:246]
        currgb[10:246] <<= 8
        delta[10:246] = 0

        colors = np.asarray(colors, np.uint8)
        index = colors[:, 0]
        colors1 = colors[:, 1:]
        colors2 = master[index]
        curpal[index] = colors2
        currgb[index] = colors2
        currgb[index] <<= 8
        delta[index] = colors1
        delta[index] -= colors2
        delta[index] <<= 8
        delta[index] /= divisor

    def fadetoblack(self, steps):
        self.palchange_countdown = steps
        self.palette_changed = True
        self.palchange_type = 1
        divisor = steps

        curpal = self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb

        currgb[10:246] = curpal[10:246]
        currgb[10:246] <<= 8
        delta[10:246] = currgb[10:246]
        delta[10:246] *= -1
        delta[10:246] <<= 8
        delta[10:246] /= divisor

    def paletteshift(self, shifts):
        self.palchange_countdown = 1
        self.palchange_type = 4
        self.palette_changed = True
        self.rgbshift_index = 0
        self.rgbshift = shifts

    def changepalette(self):
        self.bitmap_palette[10:246] = self.current_palette[10:246]
        ptype, index = self.palchange_type, self.palchange_countdown
        if ptype == 0:
            self.palette_changed = False
            self.ui.clearstate(STATE_4)
        elif ptype == 1:
            if index:
                self.stepfade()
                self.palchange_countdown = index - 1
            else:
                self.palchange_type = 0
        elif ptype == 2:
            if index:
                self.stepfade()
                index -= 1
            else:
                a, b = self.palette_from, self.palette_to
                self.delta_rgb[a:b] *= -1
                index = self.palchange_count
            self.palchange_countdown = index
        elif ptype == 3:
            index -= 1
            if not index:
                self.stepfade()
                a, b = self.palette_from, self.palette_to
                self.delta_rgb[a:b] *= -1
                index = self.palchange_count
            self.palchange_countdown = index
        elif ptype == 4:
            index -= 1
            if not index:
                rgbindex, rgbshift = self.rgbshift_index, self.rgbshift
                index, colors = rgbshift[rgbindex]
                self.current_palette[colors[:, 0]] = colors[:, 1:]
                rgbindex += 1
                if rgbindex >= len(rgbshift):
                    rgbindex = 0
                self.rgbshift_index = rgbindex
            self.palchange_countdown = index

    def stepfade(self):
        curpal = self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb
        a, b = self.palette_from, self.palette_to
        currgb[a:b] += delta[a:b]
        curpal[a:b] = np.clip(currgb[a:b] >> 8, 0, 255)

    def pausefade(self):
        self.fade_in = True
        self.palette_type = 0 # QUESTION this might be wrong
        self.palette_changed = False

    def stopfade(self):
        self.fade_in = False
        self.palette_type = 0 # QUESTION this might be wrong
        self.palette_changed = False

    def endfade(self):
        a, b = self.palette_from, self.palette_to
        if self.fade_in:
            self.current_palette[a:b] = self.master_palette[a:b]
        else:
            self.current_palette[a:b] = self.fade_color

    def resetpause(self):
        self.future_palette_changed = self.fade_in

    def torgb(self, frame):
        return self.bitmap_palette[frame.reshape(-1,)].reshape(frame.shape+(3,))

    def updatepalette(self):
        if self.palette_changed:
            self.changepalette()
            return True
        return False

    def thefutureisnow(self):
        if self.future_palette_changed:
            self.palette_changed = True
            self.future_palette_changed = False

class Shimmer(object):
    def __init__(self, type):
        self.type = type
        if type in (0, 1):
            self.value = 0
            self.counter = 8
        elif type == 2:
            self.value = 65535
            self.counter = 8
    def makeindex(self, w, h):
        # not sure this is identical, but it looks close enough
        m = self.value
        index = dissolvefilter(w, h, m)
        if self.counter:
            if self.type == 1:
                self.value = m & ~0xFFFF | m >> 1 & 0x7F7F | m << 7 & 0x8080 | 0x8008
            elif self.type == 2:
                m = m & ~0xFF00 | m >> 1 & 0x7F00
                m = m & ~0xFF | m >> 12 & 0x0F | m >> 4 & 0xF0
                self.value = m
            self.counter -= 1
        return index

try:
    from _algos import dissolvefilter
except ImportError:
    def dissolvefilter(w, h, m):
        n = m << 20 & 0xF0F00000 | m << 16 & 0x0F0F0000 | m >> 4 & 0x00000F0F | m & 0x0000F0F0
        w2 = (w+3)/4*4
        arr = np.zeros((h, w2), dtype=bool)
        for i in xrange(h-1,-1,-1):
            for j in xrange(0, w2, 4):
                arr[i, j+0] = n & 0x8000
                arr[i, j+1] = n & 0x4000
                arr[i, j+2] = n & 0x2000
                arr[i, j+3] = n & 0x8000
                n = n & ~0xFFFF | n << 8 & 0xFF00 | n >> 7 & 0xFE | n >> 15 & 0x01
            n = n >> 16 & 0xFFFF | n << 16 & ~0xFFFF
        if w != w2:
            arr = arr[:, :w]
        return arr

class MaskDict(dict):
    def __init__(self, nparr):
        self.nparr = nparr
    def __missing__(self, key):
        mask = self.nparr != key
        self[key] = mask
        return mask

def withmask(image):
    if not hasattr(image, 'nparray'):
        image.nparray = np.frombuffer(image.data, dtype=np.uint8).reshape(image.h, len(image.data) / image.h)[::-1, :image.w]
        image.masks = MaskDict(image.nparray)
    return image

