# encoding=utf_8
import re
import csv
from collections import defaultdict
from sogna.interpreter import CHAR_TEXTSUBST, CHAR_TEXTCLEAR, CHAR_TEXTCRLF, CHAR_TEXTHERO

SUBS_PATTERN = re.compile(u'(\1(.)|＾|￥|＄)|(.+)', re.DOTALL)
PERCENT_GROUP = re.compile(u'%([^%]*)%')

def extract_translatable_strings(ops):
    strings = {}
    for ins in ops:
        if ins.instruction == 0x21:
            if ins.text == CHAR_TEXTCLEAR: continue
            strings[ins.address] = ins.text
        elif ins.instruction == 0x24:
            strings[ins.address] = ins.text
        elif ins.instruction == 0x3E:
            strings[ins.address] = [c.text for c in ins.f]
        elif ins.instruction == 0x73:
            strings[ins.address] = ins.a
    return strings

def escape_translatable_string(s, names=None):
    def repl(m):
        if m.group(3):
            return m.group(3).replace('%','%%')
        elif m.group(2):
            n = ord(m.group(2))
            if names is not None: names.append(n - 1)
            return u'%{0}%'.format(n)
        elif m.group(1) == CHAR_TEXTCLEAR:
            return u'%clear%'
        elif m.group(1) == CHAR_TEXTCRLF:
            return u'%break%'
        elif m.group(1) == CHAR_TEXTHERO:
            return u'%hero%'
    return SUBS_PATTERN.sub(repl, s)

def unescape_translatable_string(s):
    def repl(m):
        q = m.group(1)
        if not q:
            return u'%'
        elif q == u'clear':
            return CHAR_TEXTCLEAR
        elif q == u'break':
            return CHAR_TEXTCRLF
        elif q == u'hero':
            return CHAR_TEXTHERO
        else:
            return CHAR_TEXTSUBST + unichr(int(q))
    return PERCENT_GROUP.sub(repl, s)

def strings2csv(strings, csvpath, wins, winorder):
    rows = []
    for winname in winorder:
        names = {}
        ops = dict((op.address, op) for op in wins[winname])
        for addr, s in sorted(strings[winname].iteritems()):
            ins = ops[addr]
            if ins.instruction == 0x21:
                snames = []
                s = escape_translatable_string(s, snames)
                row = {'string': s.encode('cp932'),
                    'names': u' / '.join(names.get(n, str(n)) for n in snames).encode('cp932')}
            elif ins.instruction == 0x24:
                names[ins.name] = s
                row = {'string': s.encode('cp932')}
            elif ins.instruction == 0x3E:
                snames = []
                ss = [escape_translatable_string(c, snames).replace('|', '||') for c in s]
                row = {'string': ' | '.join(ss).encode('cp932'),
                    'names': u' / '.join(names.get(n, str(n)) for n in snames).encode('cp932')}
            elif ins.instruction == 0x73:
                row = {'string': s.encode('cp932')}
            row['file'] = winname.encode('cp932')
            row['address'] = addr
            rows.append(row)
    with open(csvpath, 'wb') as f:
        fields = ['file', 'address', 'names', 'string']
        csv.writer(f, dialect=csv.excel_tab).writerow(fields)
        csv.DictWriter(f, fields, dialect=csv.excel_tab).writerows(rows)

def csv2strings(csvpath, wins):
    strings = defaultdict(dict)
    rows = list(csv.DictReader(open(csvpath, 'rb'), dialect=csv.excel_tab))
    ops = dict((k, dict((op.address, op) for op in v)) for k, v in wins.items())
    for row in rows:
        fn = row['file'].decode('cp932')
        if fn not in ops: continue
        addr = int(row['address'])
        s = row['string'].decode('cp932')
        ins = ops[fn][addr]
        if ins.instruction == 0x21:
            s = unescape_translatable_string(s)
        elif ins.instruction == 0x3E:
            s = [unescape_translatable_string(t.replace('||', '|')) for t in s.split(' | ')]
        strings[fn][addr] = s
    return strings

if __name__ == '__main__':
    import sys, sogna.dat, sogna.win
    datpath = sys.argv[1]
    stringspath = datpath + '.strings.txt'

    winnames = []
    wins = dict()
    strings = dict()
    with sogna.dat.DATArchive(datpath) as dat:
        for fn in dat.filenames():
            if fn.lower().endswith('.win'):
                winnames.append(fn)
                ops = sogna.win.bytecode.parse(dat.read(fn))
                wins[fn] = ops
                strings[fn] = extract_translatable_strings(ops)

    strings2csv(strings, stringspath, wins, winnames)
#
